/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */


#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include "sdk_common.h"
#include "nrf.h"
#include "nrf_error.h"
#include "nrf_esb_error_codes.h"
#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "boards.h"
#include "nrf_delay.h"
#include "app_util.h"
#define NRF_LOG_MODULE_NAME "APP"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "app_uart.h"
#include "app_timer.h"
#include "nrf_drv_clock.h"
#include "app_button.h"

#include "radio.h"
#include "gps.h"

#define NRF52840

#ifdef NRF52840
#define GPS_UART_RX 4
#define TRANSMIT_LED 14
#else
#define GPS_UART_RX 11
#define TRANSMIT_LED LED_1
#endif

#define LED1 13
#define LED2 14
#define LED3 15
#define LED4 16

#define UART_TX_BUF_SIZE 256                         //< UART TX buffer size.
#define UART_RX_BUF_SIZE 256                         //< UART RX buffer size.

// General application timer settings.
#define APP_TIMER_PRESCALER             15    // Value of the RTC1 PRESCALER register.
#define APP_TIMER_OP_QUEUE_SIZE         3     // Size of timer operation queues.
#define TRANSMIT_INTERVAL               APP_TIMER_TICKS(50, APP_TIMER_PRESCALER)    

#define NUM_OF_BUTTONS 4
#define BUTTON1 11
#define BUTTON2 12
#define BUTTON3 24
#define BUTTON4 25

static radio_modes_t mode = MODE_125_KBIT;
static int8_t tx_power = 9;

//set VDD to 3.3V
const uint32_t UICR_ADDR_0x304 __attribute__((at(0x10001304))) __attribute__((used)) = 0x5;

static bool test = true;
static char gps_test[128] = "000000.000,6325.2920,N,01026.2479,E,1,08,1.12,110.9";


const uint8_t leds_list[LEDS_NUMBER] = LEDS_LIST;

static char packet[128];

APP_TIMER_DEF(m_transmit_timer_id);

static void lfclk_request(void)
{
    uint32_t err_code = nrf_drv_clock_init();
    APP_ERROR_CHECK(err_code);
    nrf_drv_clock_lfclk_request(NULL);
}

static void transmit_timer_handler(void * p_context)
{
    if(test)
    {
        strcpy(&packet[1], gps_test);
    }
    else
    {
        get_gps_local(&packet[1]);
    }
    packet[0] = tx_power;
    
    radio_send_packet(packet, strlen(packet) + 2);
    nrf_gpio_pin_toggle(TRANSMIT_LED);
}

static void button_event_handler(uint8_t pin_no, uint8_t button_action)
{
    if(button_action == APP_BUTTON_PUSH)
    {
        switch(pin_no)
        {
            case BUTTON1:
                nrf_gpio_pin_toggle(LED4);
                if(mode == MODE_125_KBIT)
                {
                    mode = MODE_1_MBIT;
                }
                else
                {
                    mode = MODE_125_KBIT;
                }
                radio_set_mode(mode);
                break;

            case BUTTON2:
                nrf_gpio_pin_toggle(LED4);
                switch(tx_power)
                {
                    case 0:
                        tx_power = 4;
                        break;
                    case 4:
                        tx_power = 9;
                        break;
                    case 9:
                        tx_power = 0;
                        break;
                    default:
                        break;
                }
                radio_set_tx_power(tx_power);
                break;

            case BUTTON3:
                test = !test;
                break;
            
            case BUTTON4:
                break;
            
            default:
                break;
        }
    }
}

static const app_button_cfg_t app_buttons[NUM_OF_BUTTONS] = 
{
    {BUTTON1, false, BUTTON_PULL, button_event_handler},
    {BUTTON2, false, BUTTON_PULL, button_event_handler},
    {BUTTON3, false, BUTTON_PULL, button_event_handler},
    {BUTTON4, false, BUTTON_PULL, button_event_handler},
};

static void leds_init()
{
    nrf_gpio_cfg_output(LED1);
    nrf_gpio_cfg_output(LED2);
    nrf_gpio_cfg_output(LED3);
    nrf_gpio_cfg_output(LED4);
    
    nrf_gpio_pin_set(LED1);
    nrf_gpio_pin_set(LED2);
    nrf_gpio_pin_set(LED3);
    nrf_gpio_pin_set(LED4);
}

int main(void)
{
    ret_code_t err_code;
    nrf_gpio_cfg_output(TRANSMIT_LED);
    
    leds_init();
    
    lfclk_request();
    
    // Initialize the application timer module.
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
    
    err_code = app_timer_create(&m_transmit_timer_id,
                                APP_TIMER_MODE_REPEATED,
                                transmit_timer_handler);
    APP_ERROR_CHECK(err_code);
    
    err_code = app_timer_start(m_transmit_timer_id, TRANSMIT_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);
    
    //init app_button module, 10ms detection delay (button debouncing)
    err_code = app_button_init((app_button_cfg_t *)app_buttons,
                                   NUM_OF_BUTTONS,
                                   APP_TIMER_TICKS(10, APP_TIMER_PRESCALER));
    APP_ERROR_CHECK(err_code);

    err_code = app_button_enable();
    APP_ERROR_CHECK(err_code);
    
    uart_init(GPS_UART_RX);
    
    err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);
    
    radio_setup(true, tx_power, mode);
    
    while (true)
    {
        
    }
}
